#include "storage.h"

#include <EEPROM.h>
#include <esp_attr.h>

constexpr uint8_t absoluteTemperatureIndicator = 0xFF;

// The number of bytes to accumulate in RTC memory before writing it to EEPROM
constexpr uint32_t eepromChunkSize = 100;

// Storage for temperature measurements in RTC memory that survives deep sleep
RTC_DATA_ATTR uint8_t data[rtcStorageSize];
RTC_DATA_ATTR uint16_t dataIndex = 0;
RTC_DATA_ATTR uint16_t numberOfMeasurements = 0;

// The index into EEPROM storage where the next data should be written
RTC_DATA_ATTR uint32_t eepromIndex = 0;
RTC_DATA_ATTR bool isFirstRunAfterPowerOn = true;

// Keeps the last valid temperatures for each sensor
RTC_DATA_ATTR Temperature lastTemperatures[temperatureSensorCount];
RTC_DATA_ATTR uint32_t lastValidTemperatureTime[temperatureSensorCount];

bool didSetupEEPROM = false;

// On first boot, this is set to true; Afterwards it's remembered to be false
RTC_DATA_ATTR bool eepromIsConsideredEmpty = true;

constexpr uint16_t eepromOffset = sizeof(uint16_t);
constexpr uint16_t eepromDataSize = eepromSize - eepromOffset;

void setupEEPROMIfNeeded() {
    if (didSetupEEPROM) {
        return;
    }
    bool success = EEPROM.begin(eepromSize);
    if (!success) {
        Serial.println("Failed to set up EEPROM");
        didSetupEEPROM = false;
    }
    Serial.print("EEPROM size: ");
    Serial.println(eepromSize);
    didSetupEEPROM = true;
}

void resetLastMeasurements() {
    for (uint8_t index = 0; index < temperatureSensorCount; index += 1) {
        lastTemperatures[index].status == TemperatureStatus::sensorNotFound;
        lastTemperatures[index].value = 0;
        lastValidTemperatureTime[index] = 0;
    }
}

void storageConfigure() {
    if (!isFirstRunAfterPowerOn) {
        return;
    }
    isFirstRunAfterPowerOn = false;

    // Ensure that first values are stored
    resetLastMeasurements();
}

uint16_t getNumberOfBytesStoredInEEPROM() {
    if (eepromIsConsideredEmpty) {
        return 0; // Discard any previous bytes
    }
    setupEEPROMIfNeeded();
    return EEPROM.readShort(0);
}

void moveDataToEEPROMIfNeeded() {
    if (dataIndex < eepromChunkSize) {
        return; // Wait until more bytes are available
    }
    // Determine EEPROM start address
    uint16_t eepromIndex = getNumberOfBytesStoredInEEPROM() + eepromOffset;
    if (eepromIndex >= eepromDataSize) {
        return; // No more space in EEPROM, keep filling RTC memory
    }
    // Preparing to write new data, so set to false
    eepromIsConsideredEmpty = false;
    setupEEPROMIfNeeded();
    if (!didSetupEEPROM) {
        return;
    }

    // Write until EEPROM is full
    uint16_t bytesRemaining = eepromDataSize - eepromIndex;
    uint16_t bytesToWrite = min(dataIndex, bytesRemaining);
    EEPROM.writeBytes(eepromIndex, data, bytesToWrite); // TODO: Check that result == bytesToWrite
    EEPROM.writeShort(0, eepromIndex + bytesToWrite);
    EEPROM.commit();

    // Move remaining data to front of array (if any)
    uint16_t bytesToKeep = dataIndex - bytesToWrite;
    memmove(data, data + bytesToWrite, bytesToKeep);
    dataIndex -= bytesToWrite;
}

void saveByteAtCurrentIndex(uint8_t byte) {
    data[dataIndex] = byte;
    dataIndex += 1;
}

uint8_t byteForAbsoluteTemperature(long temp) {
    // Convert to temperature range
    long converted = (temp - (temperatureShiftForStorage)) / 500;
    if (converted < temperatureMinimumValue) {
        return temperatureMinimumValue;
    }
    if (converted > temperatureMaximumValue) {
        return temperatureMaximumValue;
    }
    return converted;
}

void saveTemperatureAtCurrentIndex(Temperature temp) {
    if (temp.status != TemperatureStatus::temperatureIsValid) {
        saveByteAtCurrentIndex(static_cast<uint8_t>(temp.status));
        return;
    }
    uint8_t byte = byteForAbsoluteTemperature(temp.value);
    saveByteAtCurrentIndex(byte);
}

bool needsAbsoluteTemperatureRecording(Temperature* temperatures) {
    for (uint8_t index = 0; index < temperatureSensorCount; index += 1) {
        if (temperatures[index].status != TemperatureStatus::temperatureIsValid) {
            // Error value can be encoded as differential (0x00)
            continue;
        }
        // Note: If previously only errors, then last value is 0. This also work for differential temperatures
        long diff = (lastTemperatures[index].value - temperatures[index].value) / 500 + 8;
        if (diff < 1 || diff > 15) {
            return true;
        }
    }
    return false;
}

void saveTemperatures(Temperature* temperatures) {
    if (needsAbsoluteTemperatureRecording(temperatures)) {
        // Write absolute temperatures
        saveByteAtCurrentIndex(absoluteTemperatureIndicator);
        for (uint8_t sensorIndex = 0; sensorIndex < temperatureSensorCount; sensorIndex += 1) {
            saveTemperatureAtCurrentIndex(temperatures[sensorIndex]);
            if (temperatures[sensorIndex].status == TemperatureStatus::temperatureIsValid) {
                // Only update if temperature is valid
                lastTemperatures[sensorIndex] = temperatures[sensorIndex];
                lastValidTemperatureTime[sensorIndex] = time(NULL);
            }
        }
    } else {
        // Calculate temperature differences
        uint8_t valueToStore = 0;
        for (uint8_t index = 0; index < temperatureSensorCount; index += 1) {
            uint8_t diff = 0; // Indicate sensor error
            if (temperatures[index].status == TemperatureStatus::temperatureIsValid) {
                diff = (lastTemperatures[index].value - temperatures[index].value) / 500 + 8;
            }
            if (index % 2) {
                // Store second bits
                valueToStore |= diff; // Valid range already ensured
                saveByteAtCurrentIndex(valueToStore);
                valueToStore = 0;
            } else {
                // Store in first four bits
                valueToStore = (diff << 4);
            }

            if (temperatures[index].status == TemperatureStatus::temperatureIsValid) {
                // Only update if temperature is valid
                lastTemperatures[index] = temperatures[index];
                lastValidTemperatureTime[index] = time(NULL);
            }
        }
        // Ensure storage with uneven number of sensors
        if (temperatureSensorCount % 2) {
            saveByteAtCurrentIndex(valueToStore);
        }
    }
    
    moveDataToEEPROMIfNeeded();
    numberOfMeasurements += 1;
}

uint16_t getTotalNumberOfStoredBytes() {
    return getNumberOfBytesStoredInEEPROM() + dataIndex;
}

uint16_t getNumberOfMeasurements() {
    return numberOfMeasurements;
}

uint8_t getLastTemperature(uint8_t sensorIndex) {
    if (lastTemperatures[sensorIndex].status != TemperatureStatus::temperatureIsValid) {
        return static_cast<uint8_t>(lastTemperatures[sensorIndex].status);
    }
    return byteForAbsoluteTemperature(lastTemperatures[sensorIndex].value);
}

uint16_t getTimeSinceValidTemperature(uint8_t sensorIndex) {
    return time(NULL) - lastValidTemperatureTime[sensorIndex];
}

uint16_t getRecordedBytesAtOffset(uint8_t* buffer, uint16_t offset, uint16_t count) {
    // TODO: Check limits
    uint16_t eepromByteCount = getNumberOfBytesStoredInEEPROM();
    uint16_t endIndex = offset + count;
    uint16_t eepromStart = min(offset, eepromByteCount);
    uint16_t eepromEnd = min(endIndex, eepromByteCount);
    uint16_t bytesToCopyFromEEPROM = eepromEnd - eepromStart;
    if (bytesToCopyFromEEPROM > 0) {
        EEPROM.readBytes(eepromStart, buffer, bytesToCopyFromEEPROM); // TODO: Check bytes read
        // Reduce offset and count to point to RTC memory
        offset -= bytesToCopyFromEEPROM;
        count -= bytesToCopyFromEEPROM;
    }
    offset -= eepromByteCount;
    // Copy remaining bytes from RTC memory
    uint16_t bytesToCopyFromRTC = min(count, dataIndex);
    memcpy(buffer + bytesToCopyFromEEPROM, data + offset, bytesToCopyFromRTC);
    return bytesToCopyFromEEPROM + bytesToCopyFromRTC;
}

void discardAllRecordedBytes() {
    eepromIsConsideredEmpty = true;
    dataIndex = 0;
    numberOfMeasurements = 0;
    resetLastMeasurements();
}